from .stream import App, AppSession
from .models import BaseContent, BaseMessage
from .message_types import (
    PutRoot,
    PushRight,
)
from .context import user_context, app_context, ContextModel



# AUTOGENERATED FILE
# generated by datamodel-codegen:
#   timestamp: 2025-10-12 08:23:38.742166


from __future__ import annotations

from typing import Any, List, Literal, Optional, Union

from miniappi.core.app.models import BaseContent
from pydantic import ConfigDict, Field, RootModel


class RecordStringAny(BaseContent):
    pass
    model_config = ConfigDict(
        extra='allow',
    )


Record3Cstring2Cany3E = RecordStringAny


class FormField(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    args: Optional[Record3Cstring2Cany3E] = None
    default: Optional[Any] = None
    label: str
    name: str
    type: Literal['text', 'date', 'datetime', 'boolean', 'integer', 'submit']
    value: Optional[Any] = None


class Submit(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    args: Optional[Record3Cstring2Cany3E] = None
    default: Optional[Any] = None
    label: str
    name: Literal['submit'] = 'submit'
    type: Literal['submit'] = 'submit'
    value: Optional[Any] = None


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    contentType: Literal['v0/forms/DynamicForm.vue'] = 'v0/forms/DynamicForm.vue'
    fields: List[Union[FormField, Submit]]
    id: str = Field(..., description='ID of the component')


class Root(RootModel[Props]):
    root: Props



import time
from datetime import timedelta
import asyncio
from typing import Callable, List, AsyncIterable
from contextlib import asynccontextmanager
from uuid import uuid4

from miniappi.core.stream import Streamer
from miniappi.core.stream.connection import Message
from .mock import MockClient, MockStartArgs

async def _wait_for_condition(cond: Callable, sleep=0, timeout: int=5):
    start = time.time()
    while not cond():
        if (time.time() - start) > timeout:
            raise TimeoutError(cond)
        await asyncio.sleep(sleep)

class StreamHandler:

    conn_client: MockClient
    received: List[Message]
    sent: List[Message]

    def __init__(self, stream: Streamer, request_id: str):
        self.stream = stream
        self.request_id = request_id

        self.received = []
        self.sent = []

        self._n_sent = 0
        self._channel_name = stream.url

        self.conn_client = self.stream.conn_client
        if not isinstance(self.stream.conn_client, MockClient):
            raise TypeError("Testing handler only works for MockClient")

    @property
    def channel_name(self):
        if self._channel_name is None:
            self._channel_name = str(uuid4())
        return self._channel_name

    @property
    def start_args(self):
        return MockStartArgs(
            request_id=self.request_id,
            channel=self.channel_name,
        )

    async def start_communication(self, **kwargs):
        await self.conn_client.request_queue[(self.stream.url, None)].put(
            self.start_args.model_dump()
        )

    def is_subscribed(self):
        session = self.stream.sessions.get(self.request_id)
        if not session:
            return False
        return session.subscribed

    async def wait_for_stream_ready(self, timeout=1):
        async with asyncio.timeout(timeout):
            while not self.is_subscribed():
                await asyncio.sleep(0)

    async def wait_for_sent(self):
        queue = self.conn_client.response_queue.get(
            (self.channel_name, self.request_id)
        )
        if queue is None:
            return
        while not queue.empty():
            message_data = await queue.get()
            message = Message(
                channel=self.channel_name,
                request_id=self.request_id,
                data=message_data
            )
            self.sent.append(message)
            await asyncio.sleep(0)

    async def wait_for_received(self):
        while self._n_sent != len(self.received):
            await asyncio.sleep(0)

    async def wait_for_messages(self):
        await self.wait_for_received()
        await self.wait_for_sent()

    async def get_next_sent(self):
        "Get next sent message"
        while True:
            queue = self.conn_client.response_queue.get(
                (self.channel_name, self.request_id)
            )
            if queue is not None:
                while not queue.empty():
                    message_data = await queue.get()
                    message = Message(
                        channel=self.channel_name,
                        request_id=self.request_id,
                        data=message_data
                    )
                    self.sent.append(message)
                    return message
            await asyncio.sleep(0)

    async def send_message(self, msg: dict | str):
        "Send message to the streamer"
        queue = self.conn_client.request_queue[(self.channel_name, self.request_id)]
        await queue.put(
            msg
        )
        self._n_sent += 1

@asynccontextmanager
async def listen(stream: Streamer, request_id: str = None, start_args: dict = None, wait_close=False):
    "Communicate with a stream"
    start_args = start_args or {}
    request_id = request_id or str(uuid4())

    handler = StreamHandler(
        stream=stream,
        request_id=request_id,
    )

    @stream.on_message()
    async def record_message(message: Message):
        if message.channel == handler.channel_name and message.request_id == handler.request_id:
            handler.received.append(message)

    async with asyncio.timeout(2):
        while True:
            if stream.subscribed:
                break
            await asyncio.sleep(0)

    async with asyncio.timeout(2):
        await handler.start_communication(
            **start_args
        )
        await handler.wait_for_stream_ready()

    async with asyncio.timeout(2):
        yield handler

    async with asyncio.timeout(2):
        await handler.wait_for_messages()

    if wait_close:
        async with asyncio.timeout(2):
            while stream.subscribed:
                await asyncio.sleep(0)



import asyncio
from typing import Generic, TypeVar
from collections import UserList
from miniappi.core.app import BaseContent, user_context, app_context, PushRight, AppSession
from miniappi.core.stream.connection import Message

T = TypeVar("T")

class Feed(UserList, Generic[T]):

    def __init__(self, initlist=None, id=None):
        super().__init__(initlist)
        self.id = id

    async def _push_session(self, elem, session: AppSession):
        await session.send(
            PushRight(
                id=self.id,
                data=elem
            )
        )

    async def append_all(self, element: T):
        self.data.append(element)
        for session in app_context.sessions.values():
            await self._push_session(element, session)

    async def append(self, element: T):
        self.data.append(element)
        try:
            await self._push_session(element, user_context.session)
        except LookupError:
            for session in app_context.sessions.values():
                await self._push_session(element, session)

    def as_reference(self):
        return {
            "reference": self.id,
            "data": self.data
        }



import asyncio
import logging
from typing import List, Dict, Awaitable, Any, Callable
from collections.abc import Callable
from contextlib import asynccontextmanager, AsyncExitStack
from pydantic import BaseModel
from .exceptions import CloseSessionException
from .connection import AbstractConnection, BaseStartArgs, Message

type RequestStreams = List[Callable[[dict], Awaitable[Any]]]

class StreamSession:

    callbacks_message: RequestStreams
    tasks: List[asyncio.Task]

    def __init__(self, start_conn: AbstractConnection,
                 start_args: BaseStartArgs,
                 callbacks_message: RequestStreams,
                 sessions: Dict[str, "StreamSession"]):
        self.start_conn = start_conn
        self.start_args = start_args

        self.callbacks_message = callbacks_message
        self._sessions = sessions

        self._pubsub = None

        self._sessions[start_args.request_id] = self
        self.tasks = []

        self.subscribed = False

    @property
    def request_id(self):
        return self.start_args.request_id

    async def send(self, data: dict | BaseModel):
        "Send to the response channel"

        logger = self.get_logger()
        logger.info("Sending data")
        await self._publish(data)

    @asynccontextmanager
    async def _subscribe(self):
        client = self.start_conn.from_start_args(self.start_args)
        async with client.connect() as connection:
            yield connection

    async def _listen(self):
        async for msg in self.start_conn.listen():
            yield msg

    async def _publish(self, body):
        await self.start_conn.publish(body)

    @asynccontextmanager
    async def _start(self):
        async with self._subscribe() as ws:
            self._pubsub = ws
            yield
            self._pubsub = None

    async def listen(self):
        "Listen the request channel"
        logger = self.get_logger()
        try:
            logger.info("Listening channel")
            self.subscribed = True
            async for message in self._listen():
                if message is not None:
                    await self._handle_request_message(message)
                await asyncio.sleep(0)
        finally:
            self.subscribed = False

    async def close(self, send_stop=True):
        "Close listening and remove session"
        logger = self.get_logger()
        logger.debug("Closing channel")
        #await self._pubsub.unsubscribe()
        self._sessions.pop(self.start_args.request_id)
        for task in self.tasks:
            task.cancel()
        if send_stop:
            await self._send_stop()

    async def _handle_request_message(self, msg: Message):
        if self._is_stop_message(msg):
            raise CloseSessionException("Client requested to close")
        for func in self.callbacks_message:
            await func(msg)

    def get_logger(self):
        return logging.getLogger(__name__)

    def _is_stop_message(self, message: Message):
        return message.data == "OFF"

    async def _send_stop(self):
        "Send stop message"
        await self._publish("OFF")



import asyncio
from typing import List, Dict, Tuple
from uuid import uuid4
from collections import defaultdict
from contextlib import asynccontextmanager
from miniappi.core.stream.connection import (
    BaseStartArgs,
    AbstractConnection,
    AbstractChannel,
    AbstractClient, Message
)

class MockStartArgs(BaseStartArgs):
    channel: str | None

class MockConnection(AbstractConnection):

    def __init__(self, channel: "MockChannel"):
        self.channel = channel

    async def publish(self, data: str):
        await self._store.response_queue[
            (self.channel.channel_name, self.channel.request_id)
        ].put(
            data
        )

    async def listen(self):
        while True:
            queue = self._store.request_queue
            this_queue = queue.get(
                (self.channel.channel_name, self.channel.request_id)
            )
            if this_queue is not None:
                next_msg = await this_queue.get()
                msg = Message(
                    channel=self.channel.channel_name,
                    request_id=self.channel.request_id,
                    data=next_msg
                )
                self._store.received.append(msg)
                yield msg
            await asyncio.sleep(0)

    async def listen_start(self):
        async for msg in self.listen():
            yield MockStartArgs(**msg.data)

    @property
    def _store(self):
        return self.channel.store

class MockChannel(AbstractChannel[MockConnection]):

    def __init__(self, channel: str, request_id: str | None, store: "MockClient"):
        super().__init__()
        self.channel_name = channel
        self.request_id = request_id
        self.store = store

        self.app_name = str(uuid4())
        self.app_url = f"http://localhost:0000/apps/{self.app_name}"

    @asynccontextmanager
    async def connect(self):
        yield MockConnection(channel=self)

class MockClient(AbstractClient[MockChannel]):

    def __init__(self):
        self.request_queue: Dict[Tuple[str, str], asyncio.Queue[dict]] = defaultdict(asyncio.Queue)
        self.response_queue: Dict[Tuple[str, str], asyncio.Queue[dict]] = defaultdict(asyncio.Queue)

        self.sent: List[Message] = []
        self.received: List[Message] = []
        super().__init__()

    def from_start_args(self, args: MockStartArgs):
        return MockChannel(
            channel=args.channel,
            request_id=args.request_id,
            store=self
        )

    def from_init_channel(self, channel_name: str):
        return MockChannel(
            channel=channel_name,
            request_id=None,
            store=self
        )



import asyncio
import json
import random
import datetime
import os, sys
import random
import logging
from contextlib import ExitStack, contextmanager, asynccontextmanager
from types import TracebackType
from typing import List, Dict, Any, Awaitable, Type, Generic, TypeVar, ContextManager
from collections.abc import Callable

from .models import OnMessageConfig, OnOpenConfig
from .session import StreamSession
from .exceptions import CloseSessionException, CloseStreamException
from .connection import AbstractConnection, AbstractClient, AbstractChannel, Message, BaseStartArgs, WebsocketClient
from miniappi.config import settings
from rich import print
from rich.panel import Panel

type RequestStreams = List[Callable[[dict], Awaitable[Any]]]

type CloseCallbackWithError = Callable[[type[BaseException], BaseException, TracebackType], Awaitable[Any]]
type CloseCallbackNoError = Callable[[None, None, None], Awaitable[Any]]
type CloseCallback = CloseCallbackWithError | CloseCallbackNoError

StreamSessionT = TypeVar("StreamSessionT", bound=StreamSession, default=StreamSession)
#ConnectionClientT = TypeVar("ConnectionClientT", bound=AbstractConnectionClient)

class Streamer(Generic[StreamSessionT]):

    cls_session: Type[StreamSessionT] = StreamSession

    conn_client: AbstractClient = WebsocketClient()

    def __init__(self, channel: str | None):
        self.url = channel

        self.callbacks_start = []
        self.callbacks_message: List[Callable[[Message], Awaitable[Any]]] = []
        self.callbacks_open: List[Callable[[StreamSessionT], Awaitable[Any]]] = []
        self.callbacks_close: List[CloseCallback] = []
        self.callbacks_end: List[CloseCallback] = []

        self.app_context_managers: List[ContextManager] = []
        self.channel_context_managers: List[ContextManager] = []

        self.sessions: Dict[str, StreamSessionT] = {}

        self.subscribed = False

    def get_app_context_managers(self, args: dict):
        return self.app_context_managers

    def get_channel_context_managers(self, session: StreamSessionT):
        return self.channel_context_managers

    def show_app_url(self, url: str):
        print(
            Panel(
                "Miniappi is running.\n"
                f"[bold red]App link:[/bold red] [link={url}]{url}[/link]!"
            )
        )

    @asynccontextmanager
    async def _subscribe(self, echo_link: bool | None, stack: ExitStack):
        client: AbstractChannel = self.conn_client.from_init_channel(self.url)
        async with client.connect() as connection:
            url = client.app_url
            if url is not None:
                show_url = (
                    echo_link
                    if echo_link is not None
                    else settings.echo_url
                )
                if show_url:
                    self.show_app_url(url)
            args = {
                "app_url": client.app_url,
                "name": client.app_name
            }
            for app_context in self.get_app_context_managers(args):
                stack.enter_context(app_context)
            yield connection

    async def _listen_start(self, connection: AbstractConnection):
        async for start_args in connection.listen_start():
            yield start_args

    def run(self, echo_link: bool | None = None):
        "Run app (sync)"
        asyncio.run(self.start(echo_link=echo_link))

    async def start(self, echo_link: bool | None = None):
        "Start app async"
        if self.subscribed:
            raise StreamRunningError("Stream is already subscribed")
        logger = self.get_logger("stream")

        with ExitStack() as app_stack:
            try:
                logger.debug(f"Subscribing stream: {self.url}")
                async with self._subscribe(echo_link, stack=app_stack) as start_conn:
                    self.subscribed = True
                    logger.info(f"Stream subscribed: {self.url}")
                    async with asyncio.TaskGroup() as tg:
                        await self._run_start(tg)
                        async for start_args in self._listen_start(start_conn):
                            tg.create_task(self.open_session(start_conn, start_args))
            except ExceptionGroup as exc:
                # Exception is ExceptionGroup[ExceptionGroup]
                # Check if all expected
                await self._run_end()
                is_expected = True
                for channel_exc in exc.exceptions:
                    if isinstance(channel_exc, ExceptionGroup):
                        for session_exc in channel_exc.exceptions:
                            if not isinstance(session_exc, CloseStreamException):
                                is_expected = False
                                break
                    else:
                        is_expected = False
                        break
                if not is_expected:
                    logger.exception("Stream closed unexpectedly")
                    raise
                logger.info("Stream closed expectedly")
            else:
                await self._run_end()
            finally:
                self.subscribed = False

    async def open_session(self, start_conn: AbstractConnection, start_args: BaseStartArgs):
        logger = self.get_logger("session")

        client = self.conn_client.from_start_args(start_args)
        async with client.connect() as conn:
            session = self.cls_session(
                start_conn=conn,
                start_args=start_args,
                callbacks_message=self.callbacks_message,

                sessions=self.sessions
            )

            with ExitStack() as channel_stack:
                for channel_context in self.get_channel_context_managers(session):
                    channel_stack.enter_context(channel_context)
                try:
                    async with asyncio.TaskGroup() as tg:
                        session.tasks.append(tg.create_task(session.listen()))
                        for stream in self.callbacks_open:
                            args = []
                            if stream.pass_session:
                                args.append(session)
                            session.tasks.append(tg.create_task(stream(*args)))
                        logger.info(f"Session opened for client: {start_args.request_id}")
                except ExceptionGroup as exc:
                    is_close_stream = all(isinstance(e, CloseStreamException) for e in exc.exceptions)
                    if is_close_stream:
                        logger.info("Stream closed expectedly")
                        # We raise so that stream can close itself
                        raise
                    is_expected = all(isinstance(e, CloseSessionException) for e in exc.exceptions)
                    if not is_expected:
                        logger.exception("Session closed with error")
                        raise
                    # Stream closed expectedly
                    logger.info("Session closed expectedly")
                finally:
                    await self._run_close()
                    await session.close()

    async def _run_start(self, tg: asyncio.TaskGroup):
        for cb in self.callbacks_start:
            tg.create_task(cb())

    async def _run_end(self):
        for cb in self.callbacks_end:
            await cb(*sys.exc_info())

    async def _run_close(self):
        for cb in self.callbacks_close:
            await cb(*sys.exc_info())

    def get_logger(self, name: str):
        name = f".{name}" if name else ""
        return logging.getLogger(__name__ + name)

    @contextmanager
    def temp(self):
        """Temporarily set callbacks to the app
        
        This is useful for syncing other sessions

        Examples
        --------
        ```python
        async with app.temp() as temp:

            @temp.on_open()
            async def new_users(session):
                ...

            # Callback "new_users" will called
            # now for all new user sessions
            ...

        # Callback "new_users" won't be called
        # anymore
        ...
        ```
        """
        from miniappi.flow.app import temp_app
        with temp_app(self) as t:
            yield t

    def on_message(self):
        """Callback for user sending a message (data).

        Examples
        --------
        ```python
        @app.on_message()
        async def new_message(msg):
            ...
        ```
        """
        def wrapper(func: Callable[[dict], Awaitable[Any]]):
            self.callbacks_message.append(
                OnMessageConfig(
                    func=func,
                )
            )
            return func
        return wrapper

    def on_start(self):
        """Callback for the app starting.

        Examples
        --------
        ```python
        @app.on_start()
        async def app_start():
            ...
        ```
        """
        def wrapper(func: Callable[[StreamSessionT], Awaitable[Any]]):
            self.callbacks_start.append(func)
            return func
        return wrapper

    def on_open(self, pass_session=False):
        """Callback for user opening an app
        session (user connected to the app)

        Examples
        --------
        ```python
        @app.on_open()
        async def new_user(session):
            ...
        ```
        """
        def wrapper(func: Callable[[StreamSessionT], Awaitable[Any]]):
            self.callbacks_open.append(
                OnOpenConfig(
                    func=func,
                    pass_session=pass_session
                )
            )
            return func
        return wrapper

    def on_close(self):
        """Callback for user closing the
        app session (user left the app).

        Examples
        --------
        ```python
        @app.on_close()
        async def user_left(exc_type, exc, tb):
            ...
        ```
        """
        def wrapper(func: CloseCallback) -> CloseCallback:
            self.callbacks_close.append(func)
            return func
        return wrapper

    def on_end(self):
        """Callback for app shutting down.

        Examples
        --------
        ```python
        @app.on_end()
        async def app_shutdown(exc_type, exc, tb):
            ...
        ```
        """
        def wrapper(func: CloseCallback):
            self.callbacks_end.append(func)
            return func
        return wrapper
